import os
import unittest
import numpy as np
from deepchem.utils.rdkit_utils import load_molecule
from deepchem.utils.rdkit_utils import compute_ring_center
from deepchem.utils.rdkit_utils import compute_ring_normal
from deepchem.utils.noncovalent_utils import is_pi_parallel
from deepchem.utils.noncovalent_utils import is_pi_t
from deepchem.utils.noncovalent_utils import compute_pi_stack
from deepchem.utils.noncovalent_utils import is_cation_pi
from deepchem.utils.noncovalent_utils import compute_cation_pi
from deepchem.utils.noncovalent_utils import compute_binding_pocket_cation_pi


class TestPiInteractions(unittest.TestCase):

    def setUp(self):
        current_dir = os.path.dirname(os.path.realpath(__file__))

        # simple flat ring
        from rdkit.Chem import MolFromSmiles
        from rdkit.Chem.rdDepictor import Compute2DCoords
        self.cycle4 = MolFromSmiles('C1CCC1')
        # self.cycle4.Compute2DCoords()
        Compute2DCoords(self.cycle4)

        # load and sanitize two real molecules
        _, self.prot = load_molecule(os.path.join(
            current_dir, '../../feat/tests/data/3ws9_protein_fixer_rdkit.pdb'),
                                     add_hydrogens=False,
                                     calc_charges=False,
                                     sanitize=True)

        _, self.lig = load_molecule(os.path.join(
            current_dir, '../../feat//tests/data/3ws9_ligand.sdf'),
                                    add_hydrogens=False,
                                    calc_charges=False,
                                    sanitize=True)

    def test_compute_ring_center(self):
        self.assertTrue(
            np.allclose(compute_ring_center(self.cycle4, range(4)), 0))

    def test_compute_ring_normal(self):
        normal = compute_ring_normal(self.cycle4, range(4))
        self.assertTrue(
            np.allclose(np.abs(normal / np.linalg.norm(normal)), [0, 0, 1]))

    def test_is_pi_parallel(self):
        ring1_center = np.array([0.0, 0.0, 0.0])
        ring2_center_true = np.array([4.0, 0.0, 0.0])
        ring2_center_false = np.array([10.0, 0.0, 0.0])
        ring1_normal_true = np.array([1.0, 0.0, 0.0])
        ring1_normal_false = np.array([0.0, 1.0, 0.0])

        for ring2_normal in (np.array([2.0, 0, 0]), np.array([-3.0, 0, 0])):
            # parallel normals
            self.assertTrue(
                is_pi_parallel(ring1_center, ring1_normal_true,
                               ring2_center_true, ring2_normal))
            # perpendicular normals
            self.assertFalse(
                is_pi_parallel(ring1_center, ring1_normal_false,
                               ring2_center_true, ring2_normal))
            # too far away
            self.assertFalse(
                is_pi_parallel(ring1_center, ring1_normal_true,
                               ring2_center_false, ring2_normal))

    def test_is_pi_t(self):
        ring1_center = np.array([0.0, 0.0, 0.0])
        ring2_center_true = np.array([4.0, 0.0, 0.0])
        ring2_center_false = np.array([10.0, 0.0, 0.0])
        ring1_normal_true = np.array([0.0, 1.0, 0.0])
        ring1_normal_false = np.array([1.0, 0.0, 0.0])

        for ring2_normal in (np.array([2.0, 0, 0]), np.array([-3.0, 0, 0])):
            # perpendicular normals
            self.assertTrue(
                is_pi_t(ring1_center, ring1_normal_true, ring2_center_true,
                        ring2_normal))
            # parallel normals
            self.assertFalse(
                is_pi_t(ring1_center, ring1_normal_false, ring2_center_true,
                        ring2_normal))
            # too far away
            self.assertFalse(
                is_pi_t(ring1_center, ring1_normal_true, ring2_center_false,
                        ring2_normal))

    def test_compute_pi_stack(self):
        # order of the molecules shouldn't matter
        dicts1 = compute_pi_stack(self.prot, self.lig)
        dicts2 = compute_pi_stack(self.lig, self.prot)
        for i, j in ((0, 2), (1, 3)):
            self.assertEqual(dicts1[i], dicts2[j])
            self.assertEqual(dicts1[j], dicts2[i])

        # with this criteria we should find both types of stacking
        for d in compute_pi_stack(self.lig,
                                  self.prot,
                                  dist_cutoff=7,
                                  angle_cutoff=40.):
            self.assertGreater(len(d), 0)

    def test_is_cation_pi(self):
        cation_position = np.array([[2.0, 0.0, 0.0]])
        ring_center_true = np.array([4.0, 0.0, 0.0])
        ring_center_false = np.array([10.0, 0.0, 0.0])
        ring_normal_true = np.array([1.0, 0.0, 0.0])
        ring_normal_false = np.array([0.0, 1.0, 0.0])

        # parallel normals
        self.assertTrue(
            is_cation_pi(cation_position, ring_center_true, ring_normal_true))
        # perpendicular normals
        self.assertFalse(
            is_cation_pi(cation_position, ring_center_true, ring_normal_false))
        # too far away
        self.assertFalse(
            is_cation_pi(cation_position, ring_center_false, ring_normal_true))

    # def test_compute_cation_pi(self):
    #   # TODO(rbharath): find better example, currently dicts are empty
    #   dicts1 = compute_cation_pi(self.prot, self.lig)
    #   dicts2 = compute_cation_pi(self.lig, self.prot)

    def test_compute_binding_pocket_cation_pi(self):
        # TODO find better example, currently dicts are empty
        prot_dict, lig_dict = compute_binding_pocket_cation_pi(
            self.prot, self.lig)

        exp_prot_dict, exp_lig_dict = compute_cation_pi(self.prot, self.lig)
        add_lig, add_prot = compute_cation_pi(self.lig, self.prot)
        for exp_dict, to_add in ((exp_prot_dict, add_prot), (exp_lig_dict,
                                                             add_lig)):
            for atom_idx, count in to_add.items():
                if atom_idx not in exp_dict:
                    exp_dict[atom_idx] = count
                else:
                    exp_dict[atom_idx] += count

        self.assertEqual(prot_dict, exp_prot_dict)
        self.assertEqual(lig_dict, exp_lig_dict)

    def test_compute_hydrogen_bonds(self):
        pass



import numpy as np
import tensorflow as tf
from deepchem.data import NumpyDataset
from deepchem.feat import CircularFingerprint
from deepchem.models import KerasModel
from deepchem.models.losses import HingeLoss
from tensorflow.keras.layers import Input, Dense, Dropout, Lambda


class ScScoreModel(KerasModel):
    """
    The SCScore model is a neural network model based on the work of Coley et al. [1]_ that predicts the synthetic complexity score (SCScore) of molecules and correlates it with the expected number of reaction steps required to produce the given target molecule.
    It is trained on a dataset of over 12 million reactions from the Reaxys database to impose a pairwise inequality constraint enforcing that on average the products of published chemical reactions should be more synthetically complex than their corresponding reactants.
    The learned metric (SCScore) exhibits highly desirable nonlinear behavior, particularly in recognizing increases in synthetic complexity throughout a number of linear synthetic routes.
    The SCScore model can accurately predict the synthetic complexity of a variety of molecules, including both drug-like and natural product molecules.
    SCScore has the potential to be a valuable tool for chemists who are working on drug discovery and other areas of chemistry.

    The learned metric (SCScore) exhibits highly desirable nonlinear behavior, particularly in recognizing increases in synthetic complexity throughout a number of linear synthetic routes.

    Our model uses hingeloss instead of the shifted relu loss as in the supplementary material [2]_ provided by the author.
    This could cause differentiation issues with compounds that are "close" to each other in "complexity".

    References
    ----------
    .. [1] Coley, C. W., Rogers, L., Green, W., & Jensen, K. F. (2018). "SCScore: Synthetic Complexity Learned from a Reaction Corpus". Journal of Chemical Information and Modeling, 58(2), 252-261. https://doi.org/10.1021/acs.jcim.7b00622

    .. [2] Coley, C. W., Rogers, L., Green, W., & Jensen, K. F. (2018). Supplementary material to "SCScore: Synthetic Complexity Learned from a Reaction Corpus". Journal of Chemical Information and Modeling, 58(2), 252-261. https://github.com/connorcoley/scscore
    """

    def __init__(self,
                 n_features,
                 layer_sizes=[300, 300, 300],
                 dropouts=0.0,
                 **kwargs):
        """
        Parameters
        ----------
        n_features: int
            number of features per molecule
        layer_sizes: list of int
            size of each hidden layer
        dropouts: int
            droupout to apply to each hidden layer
        kwargs
            This takes all kwards as TensorGraph
        """
        self.n_features = n_features
        self.layer_sizes = layer_sizes
        self.dropout = dropouts

        m1_features = Input(shape=(self.n_features,))
        m2_features = Input(shape=(self.n_features,))
        prev_layer1 = m1_features
        prev_layer2 = m2_features
        for layer_size in self.layer_sizes:
            layer = Dense(layer_size, activation=tf.nn.relu)
            prev_layer1 = layer(prev_layer1)
            prev_layer2 = layer(prev_layer2)
            if self.dropout > 0.0:
                prev_layer1 = Dropout(rate=self.dropout)(prev_layer1)
                prev_layer2 = Dropout(rate=self.dropout)(prev_layer2)

        readout_layer = Dense(1)
        readout_m1 = readout_layer(prev_layer1)
        readout_m2 = readout_layer(prev_layer2)
        outputs = [
            Lambda(lambda x: tf.sigmoid(x) * 4 + 1)(readout_m1),
            Lambda(lambda x: tf.sigmoid(x) * 4 + 1)(readout_m2),
            Lambda(lambda x: x[0] - x[1])([readout_m1, readout_m2])
        ]
        output_types = ['prediction', 'prediction', 'loss']
        model = tf.keras.Model(inputs=[m1_features, m2_features],
                               outputs=outputs)
        super(ScScoreModel, self).__init__(model,
                                           HingeLoss(),
                                           output_types=output_types,
                                           **kwargs)

    def default_generator(self,
                          dataset,
                          epochs=1,
                          mode='fit',
                          deterministic=True,
                          pad_batches=True):
        for epoch in range(epochs):
            for (X_b, y_b, w_b,
                 ids_b) in dataset.iterbatches(batch_size=self.batch_size,
                                               deterministic=deterministic,
                                               pad_batches=pad_batches):
                yield ([X_b[:, 0], X_b[:, 1]], [y_b], [w_b])

    def predict_mols(self, mols):
        featurizer = CircularFingerprint(size=self.n_features,
                                         radius=2,
                                         chiral=True)
        features = np.expand_dims(featurizer.featurize(mols), axis=1)
        features = np.concatenate([features, features], axis=1)
        ds = NumpyDataset(features, None, None, None)
        return self.predict(ds)[0][:, 0]



from heapq import heappush, heappushpop
from typing import List

import numpy as np

import torch.nn as nn
import torch

from deepchem.models.torch_models.layers import EncoderRNN, DecoderRNN, VariationalRandomizer
from deepchem.models.torch_models import TorchModel
from deepchem.utils.batch_utils import batch_elements, create_input_array, create_output_array


class SeqToSeq(nn.Module):
    """Implements sequence to sequence translation models.

    The model is based on the description in Sutskever et al., "Sequence to
    Sequence Learning with Neural Networks" (https://arxiv.org/abs/1409.3215),
    although this implementation uses GRUs instead of LSTMs. The goal is to
    take sequences of tokens as input, and translate each one into a different
    output sequence. The input and output sequences can both be of variable
    length, and an output sequence need not have the same length as the input
    sequence it was generated from. For example, these models were originally
    developed for use in natural language processing. In that context, the
    input might be a sequence of English words, and the output might be a
    sequence of French words. The goal would be to train the model to translate
    sentences from English to French.

    The model consists of two parts called the "encoder" and "decoder". Each one
    consists of a stack of recurrent layers. The job of the encoder is to
    transform the input sequence into a single, fixed length vector called the
    "embedding". That vector contains all relevant information from the input
    sequence. The decoder then transforms the embedding vector into the output
    sequence.

    These models can be used for various purposes. First and most obviously,
    they can be used for sequence to sequence translation. In any case where you
    have sequences of tokens, and you want to translate each one into a different
    sequence, a SeqToSeq model can be trained to perform the translation.

    Another possible use case is transforming variable length sequences into
    fixed length vectors. Many types of models require their inputs to have a
    fixed shape, which makes it difficult to use them with variable sized inputs
    (for example, when the input is a molecule, and different molecules have
    different numbers of atoms). In that case, you can train a SeqToSeq model as
    an autoencoder, so that it tries to make the output sequence identical to the
    input one. That forces the embedding vector to contain all information from
    the original sequence. You can then use the encoder for transforming
    sequences into fixed length embedding vectors, suitable to use as inputs to
    other types of models.

    Another use case is to train the decoder for use as a generative model. Here
    again you begin by training the SeqToSeq model as an autoencoder. Once
    training is complete, you can supply arbitrary embedding vectors, and
    transform each one into an output sequence. When used in this way, you
    typically train it as a variational autoencoder. This adds random noise to
    the encoder, and also adds a constraint term to the loss that forces the
    embedding vector to have a unit Gaussian distribution. You can then pick
    random vectors from a Gaussian distribution, and the output sequences should
    follow the same distribution as the training data.

    When training as a variational autoencoder, it is best to use KL cost
    annealing, as described in https://arxiv.org/abs/1511.06349. The constraint
    term in the loss is initially set to 0, so the optimizer just tries to
    minimize the reconstruction loss. Once it has made reasonable progress
    toward that, the constraint term can be gradually turned back on. The range
    of steps over which this happens is configurable.

    In this class, we establish a sequential model for the Sequence to Sequence (SeqToSeq) [1]_.

    Examples
    --------
    >>> import torch
    >>> from deepchem.models.torch_models.seqtoseq import SeqToSeq
    >>> from deepchem.utils.batch_utils import create_input_array
    >>> # Dataset of SMILES strings for testing SeqToSeq models.
    >>> train_smiles = [
    ...     'Cc1cccc(N2CCN(C(=O)C34CC5CC(CC(C5)C3)C4)CC2)c1C',
    ...     'Cn1ccnc1SCC(=O)Nc1ccc(Oc2ccccc2)cc1',
    ...     'COc1cc2c(cc1NC(=O)CN1C(=O)NC3(CCc4ccccc43)C1=O)oc1ccccc12',
    ...     'CCCc1cc(=O)nc(SCC(=O)N(CC(C)C)C2CCS(=O)(=O)C2)[nH]1',
    ... ]
    >>> tokens = set()
    >>> for s in train_smiles:
    ...     tokens = tokens.union(set(c for c in s))
    >>> token_list = sorted(list(tokens))
    >>> batch_size = len(train_smiles)
    >>> MAX_LENGTH = max(len(s) for s in train_smiles)
    >>> token_list = token_list + [" "]
    >>> input_dict = dict((x, i) for i, x in enumerate(token_list))
    >>> n_tokens = len(token_list)
    >>> embedding_dimension = 16
    >>> model = SeqToSeq(n_tokens, n_tokens, MAX_LENGTH, batch_size=batch_size,
    ...                  embedding_dimension=embedding_dimension)
    >>> inputs = create_input_array(train_smiles, MAX_LENGTH, False, batch_size,
    ...                             input_dict, " ")
    >>> output, embeddings = model([torch.tensor(inputs), torch.tensor([1])])
    >>> output.shape
    torch.Size([4, 57, 19])
    >>> embeddings.shape
    torch.Size([4, 16])

    References
    ----------
    .. [1] Sutskever et al., "Sequence to Sequence Learning with Neural Networks"

    """

    def __init__(self,
                 n_input_tokens: int,
                 n_output_tokens: int,
                 max_output_length: int,
                 encoder_layers: int = 4,
                 decoder_layers: int = 4,
                 batch_size: int = 100,
                 embedding_dimension: int = 512,
                 dropout: float = 0.0,
                 variational: bool = False,
                 annealing_start_step: int = 5000,
                 annealing_final_step: int = 10000):
        """Initialize SeqToSeq model.

        Parameters
        ----------
        n_input_tokens: int
            Number of input tokens.
        n_output_tokens: int
            Number of output tokens.
        max_output_length: int
            Maximum length of output sequence.
        encoder_layers: int (default 4)
            Number of recurrent layers in the encoder
        decoder_layers: int (default 4)
            Number of recurrent layers in the decoder
        embedding_dimension: int (default 512)
            Width of the embedding vector. This also is the width of all recurrent
            layers.
        dropout: float (default 0.0)
            Dropout probability to use during training.
        variational: bool (default False)
            If True, train the model as a variational autoencoder. This adds random
            noise to the encoder, and also constrains the embedding to follow a unit
            Gaussian distribution.
        annealing_start_step: int (default 5000)
            the step (that is, batch) at which to begin turning on the constraint
            term for KL cost annealing.
        annealing_final_step: int (default 10000)
            the tep (that is, batch) at which to finish turning on the constraint
            term for KL cost annealing.

        """
        super(SeqToSeq, self).__init__()
        self._variational = variational
        self.encoder = EncoderRNN(n_input_tokens, embedding_dimension,
                                  encoder_layers, dropout)
        self.decoder = DecoderRNN(embedding_dimension, n_output_tokens,
                                  decoder_layers, max_output_length, batch_size)
        if variational:
            self.randomizer = VariationalRandomizer(embedding_dimension,
                                                    annealing_start_step,
                                                    annealing_final_step)

    def forward(self, inputs: List):
        """Generates Embeddings using Encoder then passes it to Decoder to
        predict output sequences.

        Parameters
        ----------
        inputs: List
            List of two tensors.
            First tensor is batch of input sequence.
            Second tensor is the current global_step.

        Returns
        -------
        output: torch.Tensor
            Predicted output sequence.
        _embedding: torch.Tensor
            Embeddings generated by the Encoder.

        """
        input, global_step = inputs
        _, embedding = self.encoder(input.to(torch.long))
        self.encoder.training = False
        _, self._embedding = self.encoder(input.to(torch.long))
        self.encoder.training = True
        if self._variational:
            embedding = self.randomizer([self._embedding, global_step])
            self._embedding = self.randomizer([self._embedding, global_step],
                                              training=False)
        output, _ = self.decoder([embedding, None])
        return output, self._embedding


class SeqToSeqModel(TorchModel):
    """Implements sequence to sequence translation models.

    The model is based on the description in Sutskever et al., "Sequence to
    Sequence Learning with Neural Networks" (https://arxiv.org/abs/1409.3215),
    although this implementation uses GRUs instead of LSTMs. The goal is to
    take sequences of tokens as input, and translate each one into a different
    output sequence. The input and output sequences can both be of variable
    length, and an output sequence need not have the same length as the input
    sequence it was generated from. For example, these models were originally
    developed for use in natural language processing. In that context, the
    input might be a sequence of English words, and the output might be a
    sequence of French words. The goal would be to train the model to translate
    sentences from English to French.

    The model consists of two parts called the "encoder" and "decoder". Each one
    consists of a stack of recurrent layers. The job of the encoder is to
    transform the input sequence into a single, fixed length vector called the
    "embedding". That vector contains all relevant information from the input
    sequence. The decoder then transforms the embedding vector into the output
    sequence.

    These models can be used for various purposes. First and most obviously,
    they can be used for sequence to sequence translation. In any case where you
    have sequences of tokens, and you want to translate each one into a different
    sequence, a SeqToSeq model can be trained to perform the translation.

    Another possible use case is transforming variable length sequences into
    fixed length vectors. Many types of models require their inputs to have a
    fixed shape, which makes it difficult to use them with variable sized inputs
    (for example, when the input is a molecule, and different molecules have
    different numbers of atoms). In that case, you can train a SeqToSeq model as
    an autoencoder, so that it tries to make the output sequence identical to the
    input one. That forces the embedding vector to contain all information from
    the original sequence. You can then use the encoder for transforming
    sequences into fixed length embedding vectors, suitable to use as inputs to
    other types of models.

    Another use case is to train the decoder for use as a generative model. Here
    again you begin by training the SeqToSeq model as an autoencoder. Once
    training is complete, you can supply arbitrary embedding vectors, and
    transform each one into an output sequence. When used in this way, you
    typically train it as a variational autoencoder. This adds random noise to
    the encoder, and also adds a constraint term to the loss that forces the
    embedding vector to have a unit Gaussian distribution. You can then pick
    random vectors from a Gaussian distribution, and the output sequences should
    follow the same distribution as the training data.

    When training as a variational autoencoder, it is best to use KL cost
    annealing, as described in https://arxiv.org/abs/1511.06349. The constraint
    term in the loss is initially set to 0, so the optimizer just tries to
    minimize the reconstruction loss. Once it has made reasonable progress
    toward that, the constraint term can be gradually turned back on. The range
    of steps over which this happens is configurable.

    In this class, we establish a sequential model for the Sequence to Sequence (DTNN) [1]_.

    Examples
    --------
    >>> import torch
    >>> from deepchem.models.torch_models.seqtoseq import SeqToSeqModel
    >>> data = [
    ...     ("Cc1cccc(N2CCN(C(=O)C34CC5CC(CC(C5)C3)C4)CC2)c1C",
    ...      "Cc1cccc(N2CCN(C(=O)C34CC5CC(CC(C5)C3)C4)CC2)c1C"),
    ...     ("Cn1ccnc1SCC(=O)Nc1ccc(Oc2ccccc2)cc1",
    ...      "Cn1ccnc1SCC(=O)Nc1ccc(Oc2ccccc2)cc1"),
    ...     ("COc1cc2c(cc1NC(=O)CN1C(=O)NC3(CCc4ccccc43)C1=O)oc1ccccc12",
    ...      "COc1cc2c(cc1NC(=O)CN1C(=O)NC3(CCc4ccccc43)C1=O)oc1ccccc12"),
    ...     ("O=C1/C(=C/NC2CCS(=O)(=O)C2)c2ccccc2C(=O)N1c1ccccc1",
    ...      "O=C1/C(=C/NC2CCS(=O)(=O)C2)c2ccccc2C(=O)N1c1ccccc1"),
    ...     ("NC(=O)NC(Cc1ccccc1)C(=O)O",
    ...     "NC(=O)NC(Cc1ccccc1)C(=O)O")]
    >>> train_smiles = [s[0] for s in data]
    >>> tokens = set()
    >>> for s in train_smiles:
    ...     tokens = tokens.union(set(c for c in s))
    >>> tokens = sorted(list(tokens))
    >>> from deepchem.models.optimizers import Adam, ExponentialDecay
    >>> max_length = max(len(s) for s in train_smiles)
    >>> batch_size = 100
    >>> batches_per_epoch = len(train_smiles) / batch_size
    >>> model = SeqToSeqModel(
    ...     tokens,
    ...     tokens,
    ...     max_length,
    ...     encoder_layers=2,
    ...     decoder_layers=2,
    ...     embedding_dimension=256,
    ...     model_dir="fingerprint",
    ...     batch_size=batch_size,
    ...     learning_rate=ExponentialDecay(0.001, 0.9, batches_per_epoch))
    >>> for i in range(20):
    ...     loss = model.fit_sequences(data)
    >>> prediction = model.predict_from_sequences(train_smiles, 5)

    References
    ----------
    .. [1] Sutskever et al., "Sequence to Sequence Learning with Neural Networks"

    """

    sequence_end = object()

    def __init__(self,
                 input_tokens: List,
                 output_tokens: List,
                 max_output_length: int,
                 encoder_layers: int = 4,
                 decoder_layers: int = 4,
                 batch_size: int = 100,
                 embedding_dimension: int = 512,
                 dropout: float = 0.0,
                 reverse_input: bool = True,
                 variational: bool = False,
                 annealing_start_step: int = 5000,
                 annealing_final_step: int = 10000,
                 **kwargs):
        """Construct a SeqToSeq model.

        Parameters
        ----------
        input_tokens: list
            List of all tokens that may appear in input sequences.
        output_tokens: list
            List of all tokens that may appear in output sequences
        max_output_length: int
            Maximum length of output sequence that may be generated
        encoder_layers: int (default 4)
            Number of recurrent layers in the encoder
        decoder_layers: int (default 4)
            Number of recurrent layers in the decoder
        embedding_dimension: int (default 512)
            Width of the embedding vector. This also is the width of all
            recurrent layers.
        dropout: float (default 0.0)
            Dropout probability to use during training.
        reverse_input: bool (default True)
            If True, reverse the order of input sequences before sending
            them into the encoder. This can improve performance when
            working with long sequences.
        variational: bool (default False)
            If True, train the model as a variational autoencoder. This
            adds random noise to the encoder, and also constrains the
            embedding to follow a unit Gaussian distribution.
        annealing_start_step: int (default 5000)
            Step (that is, batch) at which to begin turning on the constraint
            term for KL cost annealing.
        annealing_final_step: int (default 10000)
            Step (that is, batch) at which to finish turning on the constraint
            term for KL cost annealing.

        """
        if SeqToSeqModel.sequence_end not in input_tokens:
            input_tokens = input_tokens + [SeqToSeqModel.sequence_end]
        if SeqToSeqModel.sequence_end not in output_tokens:
            output_tokens = output_tokens + [SeqToSeqModel.sequence_end]
        self._input_tokens = input_tokens
        self._output_tokens = output_tokens
        self._input_dict = dict((x, i) for i, x in enumerate(input_tokens))
        self._output_dict = dict((x, i) for i, x in enumerate(output_tokens))
        self._n_input_tokens = len(input_tokens)
        self._n_output_tokens = len(output_tokens)
        self._max_output_length = max_output_length
        self.batch_size = batch_size
        self._embedding_dimension = embedding_dimension
        self._dropout = dropout
        self._reverse_input = reverse_input
        self._variational = variational
        self._annealing_start_step = annealing_start_step
        self._annealing_final_step = annealing_final_step

        self.model: nn.Module = SeqToSeq(
            n_input_tokens=self._n_input_tokens,
            n_output_tokens=self._n_output_tokens,
            max_output_length=self._max_output_length,
            encoder_layers=encoder_layers,
            decoder_layers=decoder_layers,
            batch_size=self.batch_size,
            embedding_dimension=self._embedding_dimension,
            dropout=self._dropout,
            variational=self._variational,
            annealing_start_step=self._annealing_start_step,
            annealing_final_step=self._annealing_final_step)

        super(SeqToSeqModel,
              self).__init__(self.model,
                             self._create_loss(),
                             output_types=['prediction', 'embedding'],
                             batch_size=self.batch_size,
                             **kwargs)

    def _create_loss(self):
        """Create loss function for model."""
        if self._variational:
            loss = sum(self.model.randomizer.loss_list)
        else:
            loss = torch.tensor(0.0)

        def loss_fn(outputs, labels, weights):
            output = outputs[0].view(-1, outputs[0].size(-1))
            target = labels[0].view(-1)
            loss_ = nn.NLLLoss()(torch.log(output.to(torch.float32)),
                                 target.to(torch.int64))
            return loss + loss_

        return loss_fn

    def fit_sequences(self,
                      sequences: List[str],
                      max_checkpoints_to_keep: int = 5,
                      checkpoint_interval: int = 1000,
                      restore: bool = False):
        """Train this model on a set of sequences

        Parameters
        ----------
        sequences: List[str]
            Training samples to fit to. Each sample should be represented
            as a tuple of the form (input_sequence, output_sequence).
        max_checkpoints_to_keep: int
            Maximum number of checkpoints to keep. Older checkpoints are
            discarded.
        checkpoint_interval: int
            Frequency at which to write checkpoints, measured in training steps.
        restore: bool
            if True, restore the model from the most recent checkpoint and
            continue training from there. If False, retrain the model from
            scratch.

        """
        loss = self.fit_generator(
            self._generate_batches(sequences),
            max_checkpoints_to_keep=max_checkpoints_to_keep,
            checkpoint_interval=checkpoint_interval,
            restore=restore)
        return loss

    def predict_from_sequences(self, sequences: List[str], beam_width=5):
        """Given a set of input sequences, predict the output sequences.

        The prediction is done using a beam search with length normalization.

        Parameters
        ----------
        sequences: List[str]
            Input sequences to generate a prediction for
        beam_width: int (default 5)
            Beam width to use for searching.
            Set to 1 to use a simple greedy search.

        """
        result = []
        for batch in batch_elements(sequences, self.batch_size):
            features = create_input_array(batch, self._max_output_length,
                                          self._reverse_input, self.batch_size,
                                          self._input_dict,
                                          SeqToSeqModel.sequence_end)
            probs = self.predict_on_generator([
                ((features, np.array(self.get_global_step())), None, None)
            ])
            for i in range(len(batch)):
                result.append(self._beam_search(probs[i], beam_width))
        return result

    def predict_embedding(self, sequences: List[str]):  # type: ignore[override]
        """Given a set of input sequences, compute the embedding vectors.

        Parameters
        ----------
        sequences: List[str]
            Input sequences to generate embeddings for.

        """
        result = []
        for batch in batch_elements(sequences, self.batch_size):
            features = create_input_array(batch, self._max_output_length,
                                          self._reverse_input, self.batch_size,
                                          self._input_dict,
                                          SeqToSeqModel.sequence_end)
            probs = self.predict_on_generator(
                [((features, np.array(self.get_global_step())), None, None)],
                output_types=["embedding"])
            for i in range(len(batch)):
                result.append(probs[i])
        return result

    def predict_from_embedding(self,
                               embeddings: List[np.ndarray],
                               beam_width=5):
        """Given a set of embedding vectors, predict the output sequences.

        The prediction is done using a beam search with length normalization.

        Parameters
        ----------
        embeddings: List[np.ndarray]
            Embedding vectors to generate predictions for
        beam_width: int
            Beam width to use for searching.
            Set to 1 to use a simple greedy search.

        """
        result = []
        for batch in batch_elements(embeddings, self.batch_size):
            embedding_array = np.zeros(
                (self.batch_size, self._embedding_dimension), dtype=np.float32)
            for i, e in enumerate(batch):
                embedding_array[i] = e
            probs, _ = self.model.decoder(  # type: ignore
                [torch.tensor(embedding_array, device=self.device), None])
            probs = probs.cpu().detach().numpy()
            for i in range(len(batch)):
                result.append(self._beam_search(probs[i], beam_width))
        return result

    def _beam_search(self, probs: np.ndarray, beam_width: int):
        """Perform a beam search for the most likely output sequence.

        Beam search is a heuristic search algorithm often used in natural
        language processing.It is primarily used for finding the most likely
        sequence of decisions in probabilistic models. Instead of exhaustively
        exploring all possibilities, beam search maintains a set of "beams" or
        hypotheses at each step and narrows down the search space based on a
        predetermined beam width.

        Parameters
        ----------
        probs: np.ndarray
            Probability distribution over output tokens at each step.
        beam_width: int
            Beam width to use for searching.
            Set to 1 to use a simple greedy search.

        Returns
        -------
        List
            Most likely output sequence.

        """
        if beam_width == 1:
            # Do a simple greedy search.

            s = []
            for i in range(len(probs)):
                token = self._output_tokens[np.argmax(probs[i])]
                if token == SeqToSeqModel.sequence_end:
                    break
                s.append(token)
            return s

        # Do a beam search with length normalization.

        logprobs = np.log(probs)
        # Represent each candidate as (normalized prob, raw prob, sequence)
        candidates: List = [(0.0, 0.0, [])]
        for i in range(len(logprobs)):
            new_candidates: List = []
            for c in candidates:
                if len(c[2]) > 0 and c[2][-1] == SeqToSeqModel.sequence_end:
                    # This candidate sequence has already been terminated
                    if len(new_candidates) < beam_width:
                        heappush(new_candidates, c)
                    else:
                        heappushpop(new_candidates, c)
                else:
                    # Consider all possible tokens we could add to this candidate sequence.
                    for j, logprob in enumerate(logprobs[i]):
                        new_logprob = logprob + c[1]
                        newc = (new_logprob / (len(c[2]) + 1), new_logprob,
                                c[2] + [self._output_tokens[j]])
                        if len(new_candidates) < beam_width:
                            heappush(new_candidates, newc)
                        else:
                            heappushpop(new_candidates, newc)
            candidates = new_candidates
        return sorted(candidates)[-1][2][:-1]

    def _generate_batches(self, sequences: List[str]):
        """Create feed_dicts for fitting.

        Parameters
        ----------
        sequences: List[str]
            Training samples to fit to. Each sample should be represented
            as a tuple of the form (input_sequence, output_sequence).

        Returns
        -------
        Iterable
            Feed_dicts for fitting.

        """
        for batch in batch_elements(sequences, self.batch_size):
            inputs = []
            outputs = []
            for input, output in batch:
                inputs.append(input)
                outputs.append(output)
            for i in range(len(inputs), self.batch_size):
                inputs.append([])
                outputs.append([])
            features = create_input_array(inputs, self._max_output_length,
                                          self._reverse_input, self.batch_size,
                                          self._input_dict,
                                          SeqToSeqModel.sequence_end)
            labels = create_output_array(outputs, self._max_output_length,
                                         self.batch_size, self._output_dict,
                                         SeqToSeqModel.sequence_end)
            yield ([features, np.array(self.get_global_step())], [labels], [])



#!/usr/bin/env python2
# -*- coding: utf-8 -*-
"""
Created on Mon Mar  6 23:41:26 2017

@author: zqwu
"""
import numpy as np
import deepchem
from deepchem.molnet.preset_hyper_parameters import hps
from sklearn.ensemble import RandomForestClassifier
from sklearn.ensemble import RandomForestRegressor
from sklearn.linear_model import LogisticRegression
from sklearn.svm import SVC
from sklearn.kernel_ridge import KernelRidge


def benchmark_classification(train_dataset,
                             valid_dataset,
                             test_dataset,
                             tasks,
                             transformers,
                             n_features,
                             metric,
                             model,
                             test=False,
                             hyper_parameters=None,
                             seed=123):
    """Calculate performance of different models on the specific dataset & tasks

    Parameters
    ----------
    train_dataset: dataset struct
        dataset used for model training and evaluation
    valid_dataset: dataset struct
        dataset only used for model evaluation (and hyperparameter tuning)
    test_dataset: dataset struct
        dataset only used for model evaluation
    tasks: list of string
        list of targets(tasks, datasets)
    transformers: dc.trans.Transformer struct
        transformer used for model evaluation
    n_features: integer
        number of features, or length of binary fingerprints
    metric: list of dc.metrics.Metric objects
        metrics used for evaluation
    model: string,  optional
        choice of model
        'rf', 'tf', 'tf_robust', 'logreg', 'irv', 'graphconv', 'dag', 'xgb',
        'weave', 'kernelsvm', 'textcnn', 'mpnn'
    test: boolean, optional
        whether to calculate test_set performance
    hyper_parameters: dict, optional (default=None)
        hyper parameters for designated model, None = use preset values


    Returns
    -------
    train_scores : dict
    predicting results(AUC) on training set
    valid_scores : dict
    predicting results(AUC) on valid set
    test_scores : dict
    predicting results(AUC) on test set
    """
    train_scores = {}
    valid_scores = {}
    test_scores = {}

    assert model in [
        'rf', 'tf', 'tf_robust', 'logreg', 'irv', 'graphconv', 'dag', 'xgb',
        'weave', 'kernelsvm', 'textcnn', 'mpnn'
    ]
    if hyper_parameters is None:
        hyper_parameters = hps[model]
    model_name = model

    if model_name == 'tf':
        layer_sizes = hyper_parameters['layer_sizes']
        weight_init_stddevs = hyper_parameters['weight_init_stddevs']
        bias_init_consts = hyper_parameters['bias_init_consts']
        dropouts = hyper_parameters['dropouts']
        penalty = hyper_parameters['penalty']
        penalty_type = hyper_parameters['penalty_type']
        batch_size = hyper_parameters['batch_size']
        nb_epoch = hyper_parameters['nb_epoch']
        learning_rate = hyper_parameters['learning_rate']

        # Building tensorflow MultitaskDNN model
        model = deepchem.models.MultitaskClassifier(
            len(tasks),
            n_features,
            layer_sizes=layer_sizes,
            weight_init_stddevs=weight_init_stddevs,
            bias_init_consts=bias_init_consts,
            dropouts=dropouts,
            weight_decay_penalty=penalty,
            weight_decay_penalty_type=penalty_type,
            batch_size=batch_size,
            learning_rate=learning_rate,
            random_seed=seed)

    elif model_name == 'tf_robust':
        layer_sizes = hyper_parameters['layer_sizes']
        weight_init_stddevs = hyper_parameters['weight_init_stddevs']
        bias_init_consts = hyper_parameters['bias_init_consts']
        dropouts = hyper_parameters['dropouts']

        bypass_layer_sizes = hyper_parameters['bypass_layer_sizes']
        bypass_weight_init_stddevs = hyper_parameters[
            'bypass_weight_init_stddevs']
        bypass_bias_init_consts = hyper_parameters['bypass_bias_init_consts']
        bypass_dropouts = hyper_parameters['bypass_dropouts']

        penalty = hyper_parameters['penalty']
        penalty_type = hyper_parameters['penalty_type']
        batch_size = hyper_parameters['batch_size']
        nb_epoch = hyper_parameters['nb_epoch']
        learning_rate = hyper_parameters['learning_rate']

        # Building tensorflow robust MultitaskDNN model
        model = deepchem.models.RobustMultitaskClassifier(
            len(tasks),
            n_features,
            layer_sizes=layer_sizes,
            weight_init_stddevs=weight_init_stddevs,
            bias_init_consts=bias_init_consts,
            dropouts=dropouts,
            bypass_layer_sizes=bypass_layer_sizes,
            bypass_weight_init_stddevs=bypass_weight_init_stddevs,
            bypass_bias_init_consts=bypass_bias_init_consts,
            bypass_dropouts=bypass_dropouts,
            weight_decay_penalty=penalty,
            weight_decay_penalty_type=penalty_type,
            batch_size=batch_size,
            learning_rate=learning_rate,
            random_seed=seed)

    elif model_name == 'logreg':
        penalty = hyper_parameters['penalty']
        penalty_type = hyper_parameters['penalty_type']
        nb_epoch = None

        # Building scikit logistic regression model
        def model_builder(model_dir):
            sklearn_model = LogisticRegression(penalty=penalty_type,
                                               C=1. / penalty,
                                               class_weight="balanced",
                                               n_jobs=-1)
            return deepchem.models.sklearn_models.SklearnModel(
                sklearn_model, model_dir)

        model = deepchem.models.multitask.SingletaskToMultitask(
            tasks, model_builder)

    elif model_name == 'irv':
        penalty = hyper_parameters['penalty']
        batch_size = hyper_parameters['batch_size']
        nb_epoch = hyper_parameters['nb_epoch']
        learning_rate = hyper_parameters['learning_rate']
        n_K = hyper_parameters['n_K']

        # Transform fingerprints to IRV features
        transformer = deepchem.trans.IRVTransformer(n_K, len(tasks),
                                                    train_dataset)
        train_dataset = transformer.transform(train_dataset)
        valid_dataset = transformer.transform(valid_dataset)
        if test:
            test_dataset = transformer.transform(test_dataset)

        # Building tensorflow IRV model
        model = deepchem.models.TensorflowMultitaskIRVClassifier(
            len(tasks),
            K=n_K,
            penalty=penalty,
            batch_size=batch_size,
            learning_rate=learning_rate,
            random_seed=seed,
            mode='classification')

    elif model_name == 'graphconv':
        batch_size = hyper_parameters['batch_size']
        nb_epoch = hyper_parameters['nb_epoch']
        learning_rate = hyper_parameters['learning_rate']
        n_filters = hyper_parameters['n_filters']
        n_fully_connected_nodes = hyper_parameters['n_fully_connected_nodes']

        model = deepchem.models.GraphConvModel(
            len(tasks),
            graph_conv_layers=[n_filters] * 2,
            dense_layer_size=n_fully_connected_nodes,
            batch_size=batch_size,
            learning_rate=learning_rate,
            random_seed=seed,
            mode='classification')

    elif model_name == 'dag':
        batch_size = hyper_parameters['batch_size']
        nb_epoch = hyper_parameters['nb_epoch']
        learning_rate = hyper_parameters['learning_rate']
        n_graph_feat = hyper_parameters['n_graph_feat']
        default_max_atoms = hyper_parameters['default_max_atoms']

        max_atoms_train = max([mol.get_num_atoms() for mol in train_dataset.X])
        max_atoms_valid = max([mol.get_num_atoms() for mol in valid_dataset.X])
        max_atoms_test = max([mol.get_num_atoms() for mol in test_dataset.X])
        max_atoms = max([max_atoms_train, max_atoms_valid, max_atoms_test])
        max_atoms = min([max_atoms, default_max_atoms])
        print('Maximum number of atoms: %i' % max_atoms)
        reshard_size = 256
        transformer = deepchem.trans.DAGTransformer(max_atoms=max_atoms)
        train_dataset.reshard(reshard_size)
        train_dataset = transformer.transform(train_dataset)
        valid_dataset.reshard(reshard_size)
        valid_dataset = transformer.transform(valid_dataset)
        if test:
            test_dataset.reshard(reshard_size)
            test_dataset = transformer.transform(test_dataset)

        model = deepchem.models.DAGModel(len(tasks),
                                         max_atoms=max_atoms,
                                         n_atom_feat=n_features,
                                         n_graph_feat=n_graph_feat,
                                         n_outputs=30,
                                         batch_size=batch_size,
                                         learning_rate=learning_rate,
                                         random_seed=seed,
                                         use_queue=False,
                                         mode='classification')

    elif model_name == 'weave':
        batch_size = hyper_parameters['batch_size']
        nb_epoch = hyper_parameters['nb_epoch']
        learning_rate = hyper_parameters['learning_rate']
        n_graph_feat = hyper_parameters['n_graph_feat']
        n_pair_feat = hyper_parameters['n_pair_feat']

        model = deepchem.models.WeaveModel(len(tasks),
                                           n_atom_feat=n_features,
                                           n_pair_feat=n_pair_feat,
                                           n_hidden=50,
                                           n_graph_feat=n_graph_feat,
                                           batch_size=batch_size,
                                           learning_rate=learning_rate,
                                           use_queue=False,
                                           random_seed=seed,
                                           mode='classification')

    elif model_name == 'textcnn':
        batch_size = hyper_parameters['batch_size']
        nb_epoch = hyper_parameters['nb_epoch']
        learning_rate = hyper_parameters['learning_rate']
        n_embedding = hyper_parameters['n_embedding']
        filter_sizes = hyper_parameters['filter_sizes']
        num_filters = hyper_parameters['num_filters']

        all_data = deepchem.data.DiskDataset.merge(
            [train_dataset, valid_dataset, test_dataset])
        char_dict, length = deepchem.models.TextCNNModel.build_char_dict(
            all_data)

        model = deepchem.models.TextCNNModel(len(tasks),
                                             char_dict,
                                             seq_length=length,
                                             n_embedding=n_embedding,
                                             filter_sizes=filter_sizes,
                                             num_filters=num_filters,
                                             learning_rate=learning_rate,
                                             batch_size=batch_size,
                                             use_queue=False,
                                             random_seed=seed,
                                             mode='classification')

    elif model_name == 'mpnn':
        batch_size = hyper_parameters['batch_size']
        nb_epoch = hyper_parameters['nb_epoch']
        learning_rate = hyper_parameters['learning_rate']
        T = hyper_parameters['T']
        M = hyper_parameters['M']

        model = deepchem.models.MPNNModel(len(tasks),
                                          n_atom_feat=n_features[0],
                                          n_pair_feat=n_features[1],
                                          n_hidden=n_features[0],
                                          T=T,
                                          M=M,
                                          batch_size=batch_size,
                                          learning_rate=learning_rate,
                                          use_queue=False,
                                          mode="classification")

    elif model_name == 'rf':
        n_estimators = hyper_parameters['n_estimators']
        nb_epoch = None

        # Building scikit random forest model
        def model_builder(model_dir):
            sklearn_model = RandomForestClassifier(class_weight="balanced",
                                                   n_estimators=n_estimators,
                                                   n_jobs=-1)
            return deepchem.models.sklearn_models.SklearnModel(
                sklearn_model, model_dir)

        model = deepchem.models.multitask.SingletaskToMultitask(
            tasks, model_builder)

    elif model_name == 'kernelsvm':
        C = hyper_parameters['C']
        gamma = hyper_parameters['gamma']
        nb_epoch = None

        # Building scikit learn Kernel SVM model
        def model_builder(model_dir):
            sklearn_model = SVC(C=C,
                                gamma=gamma,
                                class_weight="balanced",
                                probability=True)
            return deepchem.models.SklearnModel(sklearn_model, model_dir)

        model = deepchem.models.multitask.SingletaskToMultitask(
            tasks, model_builder)

    elif model_name == 'xgb':
        max_depth = hyper_parameters['max_depth']
        learning_rate = hyper_parameters['learning_rate']
        n_estimators = hyper_parameters['n_estimators']
        gamma = hyper_parameters['gamma']
        min_child_weight = hyper_parameters['min_child_weight']
        max_delta_step = hyper_parameters['max_delta_step']
        subsample = hyper_parameters['subsample']
        colsample_bytree = hyper_parameters['colsample_bytree']
        colsample_bylevel = hyper_parameters['colsample_bylevel']
        reg_alpha = hyper_parameters['reg_alpha']
        reg_lambda = hyper_parameters['reg_lambda']
        scale_pos_weight = hyper_parameters['scale_pos_weight']
        base_score = hyper_parameters['base_score']
        seed = hyper_parameters['seed']
        early_stopping_rounds = hyper_parameters['early_stopping_rounds']
        nb_epoch = None

        esr = {'early_stopping_rounds': early_stopping_rounds}

        # Building xgboost classification model
        def model_builder(model_dir):
            import xgboost
            xgboost_model = xgboost.XGBClassifier(
                max_depth=max_depth,
                learning_rate=learning_rate,
                n_estimators=n_estimators,
                gamma=gamma,
                min_child_weight=min_child_weight,
                max_delta_step=max_delta_step,
                subsample=subsample,
                colsample_bytree=colsample_bytree,
                colsample_bylevel=colsample_bylevel,
                reg_alpha=reg_alpha,
                reg_lambda=reg_lambda,
                scale_pos_weight=scale_pos_weight,
                base_score=base_score,
                seed=seed)
            return deepchem.models.xgboost_models.XGBoostModel(
                xgboost_model, model_dir, **esr)

        model = deepchem.models.multitask.SingletaskToMultitask(
            tasks, model_builder)

    if nb_epoch is None:
        model.fit(train_dataset)
    else:
        model.fit(train_dataset, nb_epoch=nb_epoch)

    train_scores[model_name] = model.evaluate(train_dataset, metric,
                                              transformers)
    valid_scores[model_name] = model.evaluate(valid_dataset, metric,
                                              transformers)
    if test:
        test_scores[model_name] = model.evaluate(test_dataset, metric,
                                                 transformers)

    return train_scores, valid_scores, test_scores


def benchmark_regression(train_dataset,
                         valid_dataset,
                         test_dataset,
                         tasks,
                         transformers,
                         n_features,
                         metric,
                         model,
                         test=False,
                         hyper_parameters=None,
                         seed=123):
    """Calculate performance of different models on the specific dataset & tasks

    Parameters
    ----------
    train_dataset: dataset struct
        dataset used for model training and evaluation
    valid_dataset: dataset struct
        dataset only used for model evaluation (and hyperparameter tuning)
    test_dataset: dataset struct
        dataset only used for model evaluation
    tasks: list of string
        list of targets(tasks, datasets)
    transformers: dc.trans.Transformer struct
        transformer used for model evaluation
    n_features: integer
        number of features, or length of binary fingerprints
    metric: list of dc.metrics.Metric objects
        metrics used for evaluation
    model: string, optional
        choice of model
        'tf_regression', 'tf_regression_ft', 'rf_regression', 'graphconvreg',
        'dtnn', 'dag_regression', 'xgb_regression', 'weave_regression',
        'textcnn_regression', 'krr', 'ani', 'krr_ft', 'mpnn'
    test: boolean, optional
        whether to calculate test_set performance
    hyper_parameters: dict, optional (default=None)
        hyper parameters for designated model, None = use preset values


    Returns
    -------
    train_scores : dict
    predicting results(R2) on training set
    valid_scores : dict
    predicting results(R2) on valid set
    test_scores : dict
    predicting results(R2) on test set

    """
    train_scores = {}
    valid_scores = {}
    test_scores = {}

    assert model in [
        'tf_regression', 'tf_regression_ft', 'rf_regression', 'graphconvreg',
        'dtnn', 'dag_regression', 'xgb_regression', 'weave_regression',
        'textcnn_regression', 'krr', 'ani', 'krr_ft', 'mpnn'
    ]

    if hyper_parameters is None:
        hyper_parameters = hps[model]
    model_name = model

    if model_name == 'tf_regression':
        layer_sizes = hyper_parameters['layer_sizes']
        weight_init_stddevs = hyper_parameters['weight_init_stddevs']
        bias_init_consts = hyper_parameters['bias_init_consts']
        dropouts = hyper_parameters['dropouts']
        penalty = hyper_parameters['penalty']
        penalty_type = hyper_parameters['penalty_type']
        batch_size = hyper_parameters['batch_size']
        nb_epoch = hyper_parameters['nb_epoch']
        learning_rate = hyper_parameters['learning_rate']

        model = deepchem.models.MultitaskRegressor(
            len(tasks),
            n_features,
            layer_sizes=layer_sizes,
            weight_init_stddevs=weight_init_stddevs,
            bias_init_consts=bias_init_consts,
            dropouts=dropouts,
            weight_decay_penalty=penalty,
            weight_decay_penalty_type=penalty_type,
            batch_size=batch_size,
            learning_rate=learning_rate,
            seed=seed)

    elif model_name == 'tf_regression_ft':
        layer_sizes = hyper_parameters['layer_sizes']
        weight_init_stddevs = hyper_parameters['weight_init_stddevs']
        bias_init_consts = hyper_parameters['bias_init_consts']
        dropouts = hyper_parameters['dropouts']
        penalty = hyper_parameters['penalty']
        penalty_type = hyper_parameters['penalty_type']
        batch_size = hyper_parameters['batch_size']
        nb_epoch = hyper_parameters['nb_epoch']
        learning_rate = hyper_parameters['learning_rate']
        fit_transformers = [hyper_parameters['fit_transformers'](train_dataset)]

        model = deepchem.models.MultitaskFitTransformRegressor(
            n_tasks=len(tasks),
            n_features=n_features,
            layer_sizes=layer_sizes,
            weight_init_stddevs=weight_init_stddevs,
            bias_init_consts=bias_init_consts,
            dropouts=dropouts,
            weight_decay_penalty=penalty,
            weight_decay_penalty_type=penalty_type,
            batch_size=batch_size,
            learning_rate=learning_rate,
            fit_transformers=fit_transformers,
            n_eval=10,
            seed=seed)

    elif model_name == 'graphconvreg':
        batch_size = hyper_parameters['batch_size']
        nb_epoch = hyper_parameters['nb_epoch']
        learning_rate = hyper_parameters['learning_rate']
        n_filters = hyper_parameters['n_filters']
        n_fully_connected_nodes = hyper_parameters['n_fully_connected_nodes']

        model = deepchem.models.GraphConvModel(
            len(tasks),
            graph_conv_layers=[n_filters] * 2,
            dense_layer_size=n_fully_connected_nodes,
            batch_size=batch_size,
            learning_rate=learning_rate,
            random_seed=seed,
            mode='regression')

    elif model_name == 'dtnn':
        batch_size = hyper_parameters['batch_size']
        nb_epoch = hyper_parameters['nb_epoch']
        learning_rate = hyper_parameters['learning_rate']
        n_embedding = hyper_parameters['n_embedding']
        n_distance = hyper_parameters['n_distance']
        assert len(n_features) == 2, 'DTNN is only applicable to qm datasets'

        model = deepchem.models.DTNNModel(len(tasks),
                                          n_embedding=n_embedding,
                                          n_distance=n_distance,
                                          batch_size=batch_size,
                                          learning_rate=learning_rate,
                                          random_seed=seed,
                                          output_activation=False,
                                          use_queue=False,
                                          mode='regression')

    elif model_name == 'dag_regression':
        batch_size = hyper_parameters['batch_size']
        nb_epoch = hyper_parameters['nb_epoch']
        learning_rate = hyper_parameters['learning_rate']
        n_graph_feat = hyper_parameters['n_graph_feat']
        default_max_atoms = hyper_parameters['default_max_atoms']

        max_atoms_train = max([mol.get_num_atoms() for mol in train_dataset.X])
        max_atoms_valid = max([mol.get_num_atoms() for mol in valid_dataset.X])
        max_atoms_test = max([mol.get_num_atoms() for mol in test_dataset.X])
        max_atoms = max([max_atoms_train, max_atoms_valid, max_atoms_test])
        max_atoms = min([max_atoms, default_max_atoms])
        print('Maximum number of atoms: %i' % max_atoms)
        reshard_size = 256
        transformer = deepchem.trans.DAGTransformer(max_atoms=max_atoms)
        train_dataset.reshard(reshard_size)
        train_dataset = transformer.transform(train_dataset)
        valid_dataset.reshard(reshard_size)
        valid_dataset = transformer.transform(valid_dataset)
        if test:
            test_dataset.reshard(reshard_size)
            test_dataset = transformer.transform(test_dataset)

        model = deepchem.models.DAGModel(len(tasks),
                                         max_atoms=max_atoms,
                                         n_atom_feat=n_features,
                                         n_graph_feat=n_graph_feat,
                                         n_outputs=30,
                                         batch_size=batch_size,
                                         learning_rate=learning_rate,
                                         random_seed=seed,
                                         use_queue=False,
                                         mode='regression')

    elif model_name == 'weave_regression':
        batch_size = hyper_parameters['batch_size']
        nb_epoch = hyper_parameters['nb_epoch']
        learning_rate = hyper_parameters['learning_rate']
        n_graph_feat = hyper_parameters['n_graph_feat']
        n_pair_feat = hyper_parameters['n_pair_feat']

        model = deepchem.models.WeaveModel(len(tasks),
                                           n_atom_feat=n_features,
                                           n_pair_feat=n_pair_feat,
                                           n_hidden=50,
                                           n_graph_feat=n_graph_feat,
                                           batch_size=batch_size,
                                           learning_rate=learning_rate,
                                           use_queue=False,
                                           random_seed=seed,
                                           mode='regression')

    elif model_name == 'textcnn_regression':
        batch_size = hyper_parameters['batch_size']
        nb_epoch = hyper_parameters['nb_epoch']
        learning_rate = hyper_parameters['learning_rate']
        n_embedding = hyper_parameters['n_embedding']
        filter_sizes = hyper_parameters['filter_sizes']
        num_filters = hyper_parameters['num_filters']

        char_dict, length = deepchem.models.TextCNNModel.build_char_dict(
            train_dataset)

        model = deepchem.models.TextCNNModel(len(tasks),
                                             char_dict,
                                             seq_length=length,
                                             n_embedding=n_embedding,
                                             filter_sizes=filter_sizes,
                                             num_filters=num_filters,
                                             learning_rate=learning_rate,
                                             batch_size=batch_size,
                                             use_queue=False,
                                             random_seed=seed,
                                             mode='regression')

    elif model_name == 'ani':
        batch_size = hyper_parameters['batch_size']
        nb_epoch = hyper_parameters['nb_epoch']
        learning_rate = hyper_parameters['learning_rate']
        layer_structures = hyper_parameters['layer_structures']

        assert len(n_features) == 2, 'ANI is only applicable to qm datasets'
        max_atoms = n_features[0]
        atom_number_cases = np.unique(
            np.concatenate([
                train_dataset.X[:, :, 0], valid_dataset.X[:, :, 0],
                test_dataset.X[:, :, 0]
            ]))

        atom_number_cases = atom_number_cases.astype(int).tolist()
        try:
            # Remove token for paddings
            atom_number_cases.remove(0)
        except:
            pass
        ANItransformer = deepchem.trans.ANITransformer(
            max_atoms=max_atoms, atom_cases=atom_number_cases)
        train_dataset = ANItransformer.transform(train_dataset)
        valid_dataset = ANItransformer.transform(valid_dataset)
        if test:
            test_dataset = ANItransformer.transform(test_dataset)
        n_feat = ANItransformer.get_num_feats() - 1

        model = deepchem.models.ANIRegression(
            len(tasks),
            max_atoms,
            n_feat,
            layer_structures=layer_structures,
            atom_number_cases=atom_number_cases,
            batch_size=batch_size,
            learning_rate=learning_rate,
            use_queue=False,
            mode="regression",
            random_seed=seed)

    elif model_name == 'mpnn':
        batch_size = hyper_parameters['batch_size']
        nb_epoch = hyper_parameters['nb_epoch']
        learning_rate = hyper_parameters['learning_rate']
        T = hyper_parameters['T']
        M = hyper_parameters['M']

        model = deepchem.models.MPNNModel(len(tasks),
                                          n_atom_feat=n_features[0],
                                          n_pair_feat=n_features[1],
                                          n_hidden=n_features[0],
                                          T=T,
                                          M=M,
                                          batch_size=batch_size,
                                          learning_rate=learning_rate,
                                          use_queue=False,
                                          mode="regression")

    elif model_name == 'rf_regression':
        n_estimators = hyper_parameters['n_estimators']
        nb_epoch = None

        # Building scikit random forest model
        def model_builder(model_dir):
            sklearn_model = RandomForestRegressor(n_estimators=n_estimators,
                                                  n_jobs=-1)
            return deepchem.models.sklearn_models.SklearnModel(
                sklearn_model, model_dir)

        model = deepchem.models.multitask.SingletaskToMultitask(
            tasks, model_builder)

    elif model_name == 'krr':
        alpha = hyper_parameters['alpha']
        nb_epoch = None

        # Building scikit learn Kernel Ridge Regression model
        def model_builder(model_dir):
            sklearn_model = KernelRidge(kernel="rbf", alpha=alpha)
            return deepchem.models.SklearnModel(sklearn_model, model_dir)

        model = deepchem.models.multitask.SingletaskToMultitask(
            tasks, model_builder)

    elif model_name == 'krr_ft':
        alpha = hyper_parameters['alpha']
        nb_epoch = None

        ft_transformer = deepchem.trans.CoulombFitTransformer(train_dataset)
        train_dataset = ft_transformer.transform(train_dataset)
        valid_dataset = ft_transformer.transform(valid_dataset)
        test_dataset = ft_transformer.transform(test_dataset)

        # Building scikit learn Kernel Ridge Regression model
        def model_builder(model_dir):
            sklearn_model = KernelRidge(kernel="rbf", alpha=alpha)
            return deepchem.models.SklearnModel(sklearn_model, model_dir)

        model = deepchem.models.multitask.SingletaskToMultitask(
            tasks, model_builder)

    elif model_name == 'xgb_regression':
        max_depth = hyper_parameters['max_depth']
        learning_rate = hyper_parameters['learning_rate']
        n_estimators = hyper_parameters['n_estimators']
        gamma = hyper_parameters['gamma']
        min_child_weight = hyper_parameters['min_child_weight']
        max_delta_step = hyper_parameters['max_delta_step']
        subsample = hyper_parameters['subsample']
        colsample_bytree = hyper_parameters['colsample_bytree']
        colsample_bylevel = hyper_parameters['colsample_bylevel']
        reg_alpha = hyper_parameters['reg_alpha']
        reg_lambda = hyper_parameters['reg_lambda']
        scale_pos_weight = hyper_parameters['scale_pos_weight']
        base_score = hyper_parameters['base_score']
        seed = hyper_parameters['seed']
        early_stopping_rounds = hyper_parameters['early_stopping_rounds']
        nb_epoch = None

        esr = {'early_stopping_rounds': early_stopping_rounds}

        # Building xgboost regression model
        def model_builder(model_dir):
            import xgboost
            xgboost_model = xgboost.XGBRegressor(
                max_depth=max_depth,
                learning_rate=learning_rate,
                n_estimators=n_estimators,
                gamma=gamma,
                min_child_weight=min_child_weight,
                max_delta_step=max_delta_step,
                subsample=subsample,
                colsample_bytree=colsample_bytree,
                colsample_bylevel=colsample_bylevel,
                reg_alpha=reg_alpha,
                reg_lambda=reg_lambda,
                scale_pos_weight=scale_pos_weight,
                base_score=base_score,
                seed=seed)
            return deepchem.models.xgboost_models.XGBoostModel(
                xgboost_model, model_dir, **esr)

        model = deepchem.models.multitask.SingletaskToMultitask(
            tasks, model_builder)

    print('-----------------------------')
    print('Start fitting: %s' % model_name)
    if nb_epoch is None:
        model.fit(train_dataset)
    else:
        model.fit(train_dataset, nb_epoch=nb_epoch)

    train_scores[model_name] = model.evaluate(train_dataset, metric,
                                              transformers)
    valid_scores[model_name] = model.evaluate(valid_dataset, metric,
                                              transformers)
    if test:
        test_scores[model_name] = model.evaluate(test_dataset, metric,
                                                 transformers)

    return train_scores, valid_scores, test_scores


'''
def low_data_benchmark_classification(train_dataset,
                                      valid_dataset,
                                      n_features,
                                      metric,
                                      model='siamese',
                                      hyper_parameters=None,
                                      seed=123):
  """
Calculate low data benchmark performance

  Parameters
  ----------
  train_dataset : dataset struct
      loaded dataset, ConvMol struct, used for training
  valid_dataset : dataset struct
      loaded dataset, ConvMol struct, used for validation
  n_features : integer
      number of features, or length of binary fingerprints
  metric: list of dc.metrics.Metric objects
      metrics used for evaluation
  model : string,  optional (default='siamese')
      choice of which model to use, should be: siamese, attn, res
  hyper_parameters: dict, optional (default=None)
      hyper parameters for designated model, None = use preset values

  Returns
  -------
  valid_scores : dict
  predicting results(AUC) on valid set

  """
  train_scores = {}  # train set not evaluated in low data model
  valid_scores = {}

  assert model in ['siamese', 'attn', 'res']
  if hyper_parameters is None:
    hyper_parameters = hps[model]

  # Loading hyperparameters
  # num positive/negative ligands
  n_pos = hyper_parameters['n_pos']
  n_neg = hyper_parameters['n_neg']
  # Set batch sizes for network
  test_batch_size = hyper_parameters['test_batch_size']
  support_batch_size = n_pos + n_neg
  # Model structure
  n_filters = hyper_parameters['n_filters']
  n_fully_connected_nodes = hyper_parameters['n_fully_connected_nodes']

  # Traning settings
  nb_epochs = hyper_parameters['nb_epochs']
  n_train_trials = hyper_parameters['n_train_trials']
  n_eval_trials = hyper_parameters['n_eval_trials']

  learning_rate = hyper_parameters['learning_rate']

  tf.set_random_seed(seed)
  support_graph = deepchem.nn.SequentialSupportGraph(n_features)
  prev_features = n_features
  for count, n_filter in enumerate(n_filters):
    support_graph.add(
        deepchem.nn.GraphConv(int(n_filter), prev_features, activation='relu'))
    support_graph.add(deepchem.nn.GraphPool())
    prev_features = int(n_filter)

  for count, n_fcnode in enumerate(n_fully_connected_nodes):
    support_graph.add(
        deepchem.nn.Dense(int(n_fcnode), prev_features, activation='tanh'))
    prev_features = int(n_fcnode)

  support_graph.add_test(
      deepchem.nn.GraphGather(test_batch_size, activation='tanh'))
  support_graph.add_support(
      deepchem.nn.GraphGather(support_batch_size, activation='tanh'))
  if model in ['siamese']:
    pass
  elif model in ['attn']:
    max_depth = hyper_parameters['max_depth']
    support_graph.join(
        deepchem.nn.AttnLSTMEmbedding(test_batch_size, support_batch_size,
                                      prev_features, max_depth))
  elif model in ['res']:
    max_depth = hyper_parameters['max_depth']
    support_graph.join(
        deepchem.nn.ResiLSTMEmbedding(test_batch_size, support_batch_size,
                                      prev_features, max_depth))

  model_low_data = deepchem.models.SupportGraphClassifier(
      support_graph,
      test_batch_size=test_batch_size,
      support_batch_size=support_batch_size,
      learning_rate=learning_rate)

  print('-------------------------------------')
  print('Start fitting by low data model: ' + model)
  # Fit trained model
  model_low_data.fit(
      train_dataset,
      nb_epochs=nb_epochs,
      n_episodes_per_epoch=n_train_trials,
      n_pos=n_pos,
      n_neg=n_neg,
      log_every_n_samples=50)

  # Evaluating low data model
  valid_scores[model] = model_low_data.evaluate(
      valid_dataset, metric, n_pos, n_neg, n_trials=n_eval_trials)

  return valid_scores
'''



"""
clearance dataset loader.
"""
import os
import deepchem as dc
from deepchem.molnet.load_function.molnet_loader import TransformerGenerator, _MolnetLoader
from deepchem.data import Dataset
from typing import List, Optional, Tuple, Union

CLEARANCE_URL = "https://deepchemdata.s3-us-west-1.amazonaws.com/datasets/clearance.csv"
CLEARANCE_TASKS = ['target']


class _ClearanceLoader(_MolnetLoader):

    def create_dataset(self) -> Dataset:
        dataset_file = os.path.join(self.data_dir, "clearance.csv")
        if not os.path.exists(dataset_file):
            dc.utils.data_utils.download_url(url=CLEARANCE_URL,
                                             dest_dir=self.data_dir)
        loader = dc.data.CSVLoader(tasks=self.tasks,
                                   feature_field="smiles",
                                   featurizer=self.featurizer)
        return loader.create_dataset(dataset_file, shard_size=8192)


def load_clearance(
    featurizer: Union[dc.feat.Featurizer, str] = 'ECFP',
    splitter: Union[dc.splits.Splitter, str, None] = 'scaffold',
    transformers: List[Union[TransformerGenerator, str]] = ['log'],
    reload: bool = True,
    data_dir: Optional[str] = None,
    save_dir: Optional[str] = None,
    **kwargs
) -> Tuple[List[str], Tuple[Dataset, ...], List[dc.trans.Transformer]]:
    """
    Load clearance datasets.

    Parameters
    ----------
    featurizer: Featurizer or str
        the featurizer to use for processing the data.  Alternatively you can pass
        one of the names from dc.molnet.featurizers as a shortcut.
    splitter: Splitter or str
        the splitter to use for splitting the data into training, validation, and
        test sets.  Alternatively you can pass one of the names from
        dc.molnet.splitters as a shortcut.  If this is None, all the data
        will be included in a single dataset.
    transformers: list of TransformerGenerators or strings
        the Transformers to apply to the data.  Each one is specified by a
        TransformerGenerator or, as a shortcut, one of the names from
        dc.molnet.transformers.
    reload: bool
        if True, the first call for a particular featurizer and splitter will cache
        the datasets to disk, and subsequent calls will reload the cached datasets.
    data_dir: str
        a directory to save the raw data in
    save_dir: str
        a directory to save the dataset in
    """
    loader = _ClearanceLoader(featurizer, splitter, transformers,
                              CLEARANCE_TASKS, data_dir, save_dir, **kwargs)
    return loader.load_dataset('clearance', reload)



"""
Calculated formation energies for inorganic crystals from Materials Project.
"""
import os
import deepchem as dc
from deepchem.molnet.load_function.molnet_loader import TransformerGenerator, _MolnetLoader
from deepchem.data import Dataset
from typing import List, Optional, Tuple, Union

MPFORME_URL = 'https://deepchemdata.s3-us-west-1.amazonaws.com/datasets/mp_formation_energy.tar.gz'
MPFORME_TASKS = ['formation_energy']


class _MPFormationLoader(_MolnetLoader):

    def create_dataset(self) -> Dataset:
        dataset_file = os.path.join(self.data_dir, 'mp_formation_energy.json')
        targz_file = os.path.join(self.data_dir, 'mp_formation_energy.tar.gz')
        if not os.path.exists(dataset_file):
            if not os.path.exists(targz_file):
                dc.utils.data_utils.download_url(url=MPFORME_URL,
                                                 dest_dir=self.data_dir)
            dc.utils.data_utils.untargz_file(targz_file, self.data_dir)
        loader = dc.data.JsonLoader(tasks=self.tasks,
                                    feature_field="structure",
                                    label_field="formation_energy",
                                    featurizer=self.featurizer)
        return loader.create_dataset(dataset_file)


def load_mp_formation_energy(
    featurizer: Union[dc.feat.Featurizer, str] = dc.feat.SineCoulombMatrix(),
    splitter: Union[dc.splits.Splitter, str, None] = 'random',
    transformers: List[Union[TransformerGenerator, str]] = ['normalization'],
    reload: bool = True,
    data_dir: Optional[str] = None,
    save_dir: Optional[str] = None,
    **kwargs
) -> Tuple[List[str], Tuple[Dataset, ...], List[dc.trans.Transformer]]:
    """Load mp formation energy dataset.

    Contains 132752 calculated formation energies and inorganic
    crystal structures from the Materials Project database. In benchmark
    studies, random forest models achieved a mean average error of
    0.116 eV/atom during five-folded nested cross validation on this
    dataset.

    For more details on the dataset see [1]_. For more details
    on previous benchmarks for this dataset, see [2]_.

    Parameters
    ----------
    featurizer: Featurizer or str
        the featurizer to use for processing the data.  Alternatively you can pass
        one of the names from dc.molnet.featurizers as a shortcut.
    splitter: Splitter or str
        the splitter to use for splitting the data into training, validation, and
        test sets.  Alternatively you can pass one of the names from
        dc.molnet.splitters as a shortcut.  If this is None, all the data
        will be included in a single dataset.
    transformers: list of TransformerGenerators or strings
        the Transformers to apply to the data.  Each one is specified by a
        TransformerGenerator or, as a shortcut, one of the names from
        dc.molnet.transformers.
    reload: bool
        if True, the first call for a particular featurizer and splitter will cache
        the datasets to disk, and subsequent calls will reload the cached datasets.
    data_dir: str
        a directory to save the raw data in
    save_dir: str
        a directory to save the dataset in

    Returns
    -------
    tasks, datasets, transformers : tuple
        tasks : list
            Column names corresponding to machine learning target variables.
        datasets : tuple
            train, validation, test splits of data as
            ``deepchem.data.datasets.Dataset`` instances.
        transformers : list
            ``deepchem.trans.transformers.Transformer`` instances applied
            to dataset.

    References
    ----------
    .. [1] A. Jain*, S.P. Ong*, et al. (*=equal contributions) The Materials Project:
        A materials genome approach to accelerating materials innovation APL Materials,
        2013, 1(1), 011002. doi:10.1063/1.4812323 (2013).
    .. [2] Dunn, A. et al. "Benchmarking Materials Property Prediction Methods: The Matbench
        Test Set and Automatminer Reference Algorithm." https://arxiv.org/abs/2005.00707 (2020)

    Examples
    --------
    >>>
    >> import deepchem as dc
    >> tasks, datasets, transformers = dc.molnet.load_mp_formation_energy()
    >> train_dataset, val_dataset, test_dataset = datasets
    >> n_tasks = len(tasks)
    >> n_features = train_dataset.get_data_shape()[0]
    >> model = dc.models.MultitaskRegressor(n_tasks, n_features)

    """
    loader = _MPFormationLoader(featurizer, splitter, transformers,
                                MPFORME_TASKS, data_dir, save_dir, **kwargs)
    return loader.load_dataset('mp-forme', reload)



# flake8: noqa
from deepchem.hyper.base_classes import HyperparamOpt
from deepchem.hyper.grid_search import GridHyperparamOpt
from deepchem.hyper.gaussian_process import GaussianProcessHyperparamOpt
from deepchem.hyper.random_search import RandomHyperparamOpt



"""Model-Agnostic Meta-Learning (MAML) algorithm for low data learning."""

import os
import shutil
import tempfile
import time

import tensorflow as tf

from deepchem.models.optimizers import Adam, GradientDescent


class MAML(object):
    """Implements the Model-Agnostic Meta-Learning algorithm for low data learning.

    The algorithm is described in Finn et al., "Model-Agnostic Meta-Learning for Fast
    Adaptation of Deep Networks" (https://arxiv.org/abs/1703.03400).  It is used for
    training models that can perform a variety of tasks, depending on what data they
    are trained on.  It assumes you have training data for many tasks, but only a small
    amount for each one.  It performs "meta-learning" by looping over tasks and trying
    to minimize the loss on each one *after* one or a few steps of gradient descent.
    That is, it does not try to create a model that can directly solve the tasks, but
    rather tries to create a model that is very easy to train.

    To use this class, create a subclass of MetaLearner that encapsulates the model
    and data for your learning problem.  Pass it to a MAML object and call fit().
    You can then use train_on_current_task() to fine tune the model for a particular
    task.
    """

    def __init__(self,
                 learner,
                 learning_rate=0.001,
                 optimization_steps=1,
                 meta_batch_size=10,
                 optimizer=Adam(),
                 model_dir=None):
        """Create an object for performing meta-optimization.

        Parameters
        ----------
        learner: MetaLearner
            defines the meta-learning problem
        learning_rate: float or Tensor
            the learning rate to use for optimizing each task (not to be confused with the one used
            for meta-learning).  This can optionally be made a variable (represented as a
            Tensor), in which case the learning rate will itself be learnable.
        optimization_steps: int
            the number of steps of gradient descent to perform for each task
        meta_batch_size: int
            the number of tasks to use for each step of meta-learning
        optimizer: Optimizer
            the optimizer to use for meta-learning (not to be confused with the gradient descent
            optimization performed for each task)
        model_dir: str
            the directory in which the model will be saved.  If None, a temporary directory will be created.
        """
        # Record inputs.

        self.learner = learner
        self.learning_rate = learning_rate
        self.optimization_steps = optimization_steps
        self.meta_batch_size = meta_batch_size
        self.optimizer = optimizer

        # Create the output directory if necessary.

        self._model_dir_is_temp = False
        if model_dir is not None:
            if not os.path.exists(model_dir):
                os.makedirs(model_dir)
        else:
            model_dir = tempfile.mkdtemp()
            self._model_dir_is_temp = True
        self.model_dir = model_dir
        self.save_file = "%s/%s" % (self.model_dir, "model")

        # Create the optimizers for meta-optimization and task optimization.

        self._global_step = tf.Variable(0, trainable=False)
        self._tf_optimizer = optimizer._create_tf_optimizer(self._global_step)
        task_optimizer = GradientDescent(learning_rate=self.learning_rate)
        self._tf_task_optimizer = task_optimizer._create_tf_optimizer(
            self._global_step)

        # Create a Checkpoint for saving.

        self._checkpoint = tf.train.Checkpoint()
        self._checkpoint.listed = learner.variables

    def __del__(self):
        if '_model_dir_is_temp' in dir(self) and self._model_dir_is_temp:
            shutil.rmtree(self.model_dir)

    def fit(self,
            steps,
            max_checkpoints_to_keep=5,
            checkpoint_interval=600,
            restore=False):
        """Perform meta-learning to train the model.

        Parameters
        ----------
        steps: int
            the number of steps of meta-learning to perform
        max_checkpoints_to_keep: int
            the maximum number of checkpoint files to keep.  When this number is reached, older
            files are deleted.
        checkpoint_interval: float
            the time interval at which to save checkpoints, measured in seconds
        restore: bool
            if True, restore the model from the most recent checkpoint before training
            it further
        """
        if restore:
            self.restore()
        manager = tf.train.CheckpointManager(self._checkpoint, self.model_dir,
                                             max_checkpoints_to_keep)
        checkpoint_time = time.time()

        # Main optimization loop.

        learner = self.learner
        variables = learner.variables
        for i in range(steps):
            for j in range(self.meta_batch_size):
                learner.select_task()
                meta_loss, meta_gradients = self._compute_meta_loss(
                    learner.get_batch(), learner.get_batch(), variables)
                if j == 0:
                    summed_gradients = meta_gradients
                else:
                    summed_gradients = [
                        s + g for s, g in zip(summed_gradients, meta_gradients)
                    ]
            self._tf_optimizer.apply_gradients(zip(summed_gradients, variables))

            # Do checkpointing.

            if i == steps - 1 or time.time(
            ) >= checkpoint_time + checkpoint_interval:
                manager.save()
                checkpoint_time = time.time()

    @tf.function
    def _compute_meta_loss(self, inputs, inputs2, variables):
        """This is called during fitting to compute the meta-loss (the loss after a
        few steps of optimization), and its gradient.
        """
        updated_variables = variables
        with tf.GradientTape() as meta_tape:
            for k in range(self.optimization_steps):
                with tf.GradientTape() as tape:
                    loss, _ = self.learner.compute_model(
                        inputs, updated_variables, True)
                gradients = tape.gradient(loss, updated_variables)
                updated_variables = [
                    v if g is None else v - self.learning_rate * g
                    for v, g in zip(updated_variables, gradients)
                ]
            meta_loss, _ = self.learner.compute_model(inputs2,
                                                      updated_variables, True)
        meta_gradients = meta_tape.gradient(meta_loss, variables)
        return meta_loss, meta_gradients

    def restore(self):
        """Reload the model parameters from the most recent checkpoint file."""
        last_checkpoint = tf.train.latest_checkpoint(self.model_dir)
        if last_checkpoint is None:
            raise ValueError('No checkpoint found')
        self._checkpoint.restore(last_checkpoint)

    def train_on_current_task(self, optimization_steps=1, restore=True):
        """Perform a few steps of gradient descent to fine tune the model on the current task.

        Parameters
        ----------
        optimization_steps: int
            the number of steps of gradient descent to perform
        restore: bool
            if True, restore the model from the most recent checkpoint before optimizing
        """
        if restore:
            self.restore()
        variables = self.learner.variables
        for i in range(optimization_steps):
            inputs = self.learner.get_batch()
            with tf.GradientTape() as tape:
                loss, _ = self.learner.compute_model(inputs, variables, True)
            gradients = tape.gradient(loss, variables)
            self._tf_task_optimizer.apply_gradients(zip(gradients, variables))

    def predict_on_batch(self, inputs):
        """Compute the model's outputs for a batch of inputs.

        Parameters
        ----------
        inputs: list of arrays
            the inputs to the model

        Returns
        -------
        (loss, outputs) where loss is the value of the model's loss function, and
        outputs is a list of the model's outputs
        """
        return self.learner.compute_model(inputs, self.learner.variables, False)



"""
SPLIF Fingerprints for molecular complexes.
"""
import logging
import itertools
import numpy as np
from deepchem.utils.hash_utils import hash_ecfp_pair
from deepchem.utils.rdkit_utils import load_complex
from deepchem.utils.rdkit_utils import compute_all_ecfp
from deepchem.utils.rdkit_utils import MoleculeLoadException
from deepchem.utils.rdkit_utils import compute_contact_centroid
from deepchem.feat import ComplexFeaturizer
from deepchem.utils.hash_utils import vectorize
from deepchem.utils.voxel_utils import voxelize
from deepchem.utils.voxel_utils import convert_atom_pair_to_voxel
from deepchem.utils.geometry_utils import compute_pairwise_distances
from deepchem.utils.geometry_utils import subtract_centroid

from typing import Optional, Tuple, Dict, List

logger = logging.getLogger(__name__)

SPLIF_CONTACT_BINS = [(0, 2.0), (2.0, 3.0), (3.0, 4.5)]


def compute_splif_features_in_range(frag1: Tuple,
                                    frag2: Tuple,
                                    pairwise_distances: np.ndarray,
                                    contact_bin: List,
                                    ecfp_degree: int = 2) -> Dict:
    """Computes SPLIF features for close atoms in molecular complexes.

    Finds all frag1 atoms that are > contact_bin[0] and <
    contact_bin[1] away from frag2 atoms. Then, finds the ECFP
    fingerprints for the contacting atoms. Returns a dictionary
    mapping (frag1_index_i, frag2_index_j) --> (frag1_ecfp_i,
    frag2_ecfp_j)

    Parameters
    ----------
    frag1: Tuple
        A tuple of (coords, mol) returned by `load_molecule`.
    frag2: Tuple
        A tuple of (coords, mol) returned by `load_molecule`.
    contact_bins: np.ndarray
        Ranges of pair distances which are placed in separate bins.
    pairwise_distances: np.ndarray
        Array of pairwise fragment-fragment distances (Angstroms)
    ecfp_degree: int
        ECFP radius
    """
    contacts = np.nonzero((pairwise_distances > contact_bin[0]) &
                          (pairwise_distances < contact_bin[1]))
    frag1_atoms = set([int(c) for c in contacts[0].tolist()])

    frag1_ecfp_dict = compute_all_ecfp(frag1[1],
                                       indices=frag1_atoms,
                                       degree=ecfp_degree)
    frag2_ecfp_dict = compute_all_ecfp(frag2[1], degree=ecfp_degree)
    splif_dict = {
        contact: (frag1_ecfp_dict[contact[0]], frag2_ecfp_dict[contact[1]])
        for contact in zip(contacts[0], contacts[1])
    }
    return splif_dict


def featurize_splif(frag1, frag2, contact_bins, pairwise_distances,
                    ecfp_degree):
    """Computes SPLIF featurization of fragment interactions binding pocket.

    For each contact range (i.e. 1 A to 2 A, 2 A to 3 A, etc.)
    compute a dictionary mapping (frag1_index_i, frag2_index_j)
    tuples --> (frag1_ecfp_i, frag2_ecfp_j) tuples. Return a
    list of such splif dictionaries.

    Parameters
    ----------
    frag1: Tuple
        A tuple of (coords, mol) returned by `load_molecule`.
    frag2: Tuple
        A tuple of (coords, mol) returned by `load_molecule`.
    contact_bins: np.ndarray
        Ranges of pair distances which are placed in separate bins.
    pairwise_distances: np.ndarray
        Array of pairwise fragment-fragment distances (Angstroms)
    ecfp_degree: int
        ECFP radius, the graph distance at which fragments are computed.

    Returns
    -------
    Dictionaries of SPLIF interactions suitable for `vectorize` or
    `voxelize`.
    """
    splif_dicts = []
    for i, contact_bin in enumerate(contact_bins):
        splif_dicts.append(
            compute_splif_features_in_range(frag1, frag2, pairwise_distances,
                                            contact_bin, ecfp_degree))

    return splif_dicts


class SplifFingerprint(ComplexFeaturizer):
    """Computes SPLIF Fingerprints for a macromolecular complex.

    SPLIF fingerprints are based on a technique introduced in the
    following paper.

    Da, C., and D. Kireev. "Structural protein–ligand interaction
    fingerprints (SPLIF) for structure-based virtual screening:
    method and benchmark study." Journal of chemical information
    and modeling 54.9 (2014): 2555-2561.

    SPLIF fingerprints are a subclass of `ComplexFeaturizer`. It
    requires 3D coordinates for a molecular complex. For each ligand
    atom, it identifies close pairs of atoms from different molecules.
    These atom pairs are expanded to 2D circular fragments and a
    fingerprint for the union is turned on in the bit vector. Note that
    we slightly generalize the original paper by not requiring the
    interacting molecules to be proteins or ligands.

    This is conceptually pretty similar to
    `ContactCircularFingerprint` but computes ECFP fragments only
    for direct contacts instead of the entire contact region.

    For a macromolecular complex, returns a vector of shape
    `(len(contact_bins)*size,)`
    """

    def __init__(self, contact_bins=None, radius=2, size=8):
        """
        Parameters
        ----------
        contact_bins: list[tuple]
            List of contact bins. If not specified is set to default
            `[(0, 2.0), (2.0, 3.0), (3.0, 4.5)]`.
        radius : int, optional (default 2)
            Fingerprint radius used for circular fingerprints.
        size: int, optional (default 8)
            Length of generated bit vector.
        """
        if contact_bins is None:
            self.contact_bins = SPLIF_CONTACT_BINS
        else:
            self.contact_bins = contact_bins
        self.size = size
        self.radius = radius

    def _featurize(self, datapoint, **kwargs):
        """
        Compute featurization for a molecular complex

        Parameters
        ----------
        datapoint: Tuple[str, str]
            Filenames for molecule and protein.
        """
        if 'complex' in kwargs:
            datapoint = kwargs.get("complex")
            raise DeprecationWarning(
                'Complex is being phased out as a parameter, please pass "datapoint" instead.'
            )

        try:
            fragments = load_complex(datapoint, add_hydrogens=False)

        except MoleculeLoadException:
            logger.warning(
                "This molecule cannot be loaded by Rdkit. Returning None")
            return None
        pairwise_features = []
        # We compute pairwise contact fingerprints
        for (frag1, frag2) in itertools.combinations(fragments, 2):
            # Get coordinates
            distances = compute_pairwise_distances(frag1[0], frag2[0])
            # distances = compute_pairwise_distances(prot_xyz, lig_xyz)
            vectors = [
                vectorize(hash_ecfp_pair,
                          feature_dict=splif_dict,
                          size=self.size)
                for splif_dict in featurize_splif(
                    frag1, frag2, self.contact_bins, distances, self.radius)
            ]
            pairwise_features += vectors
        pairwise_features = np.concatenate(pairwise_features)
        return pairwise_features


class SplifVoxelizer(ComplexFeaturizer):
    """Computes SPLIF voxel grid for a macromolecular complex.

    SPLIF fingerprints are based on a technique introduced in the
    following paper [1]_.

    The SPLIF voxelizer localizes local SPLIF descriptors in
    space, by assigning features to the voxel in which they
    originated. This technique may be useful for downstream
    learning methods such as convolutional networks.

    Featurizes a macromolecular complex into a tensor of shape
    `(voxels_per_edge, voxels_per_edge, voxels_per_edge, size)`
    where `voxels_per_edge = int(box_width/voxel_width)`.

    References
    ----------
    .. [1] Da, C., and D. Kireev. "Structural protein–ligand interaction
    fingerprints (SPLIF) for structure-based virtual screening:
    method and benchmark study." Journal of chemical information
    and modeling 54.9 (2014): 2555-2561.
    """

    def __init__(self,
                 cutoff: float = 4.5,
                 contact_bins: Optional[List] = None,
                 radius: int = 2,
                 size: int = 8,
                 box_width: float = 16.0,
                 voxel_width: float = 1.0):
        """
        Parameters
        ----------
        cutoff: float (default 4.5)
            Distance cutoff in angstroms for molecules in complex.
        contact_bins: list[tuple]
            List of contact bins. If not specified is set to default
            `[(0, 2.0), (2.0, 3.0), (3.0, 4.5)]`.
        radius : int, optional (default 2)
            Fingerprint radius used for circular fingerprints.
        size: int, optional (default 8)
            Length of generated bit vector.
        box_width: float, optional (default 16.0)
            Size of a box in which voxel features are calculated. Box
            is centered on a ligand centroid.
        voxel_width: float, optional (default 1.0)
            Size of a 3D voxel in a grid.
        """
        self.cutoff = cutoff
        if contact_bins is None:
            self.contact_bins = SPLIF_CONTACT_BINS
        else:
            self.contact_bins = contact_bins
        self.size = size
        self.radius = radius
        self.box_width = box_width
        self.voxel_width = voxel_width
        self.voxels_per_edge = int(self.box_width / self.voxel_width)

    def _featurize(self, datapoint, **kwargs):
        """
        Compute featurization for a molecular complex

        Parameters
        ----------
        datapoint: Tuple[str, str]
            Filenames for molecule and protein.
        """
        if 'complex' in kwargs:
            datapoint = kwargs.get("complex")
            raise DeprecationWarning(
                'Complex is being phased out as a parameter, please pass "datapoint" instead.'
            )

        try:
            fragments = load_complex(datapoint, add_hydrogens=False)

        except MoleculeLoadException:
            logger.warning(
                "This molecule cannot be loaded by Rdkit. Returning None")
            return None
        pairwise_features = []
        # We compute pairwise contact fingerprints
        centroid = compute_contact_centroid(fragments, cutoff=self.cutoff)
        for (frag1, frag2) in itertools.combinations(fragments, 2):
            distances = compute_pairwise_distances(frag1[0], frag2[0])
            frag1_xyz = subtract_centroid(frag1[0], centroid)
            frag2_xyz = subtract_centroid(frag2[0], centroid)
            xyzs = [frag1_xyz, frag2_xyz]
            pairwise_features.append(
                np.concatenate([
                    voxelize(convert_atom_pair_to_voxel,
                             hash_function=hash_ecfp_pair,
                             coordinates=xyzs,
                             box_width=self.box_width,
                             voxel_width=self.voxel_width,
                             feature_dict=splif_dict,
                             nb_channel=self.size)
                    for splif_dict in featurize_splif(
                        frag1, frag2, self.contact_bins, distances, self.radius)
                ],
                               axis=-1))
        # Features are of shape (voxels_per_edge, voxels_per_edge, voxels_per_edge, 1) so we should concatenate on the last axis.
        return np.concatenate(pairwise_features, axis=-1)



import numpy as np
from flaky import flaky
import pytest
try:
    import tensorflow as tf
    from tensorflow.keras import layers
    has_tensorflow = True
except:
    has_tensorflow = False

import deepchem as dc
from deepchem.data import NumpyDataset


@flaky
@pytest.mark.tensorflow
def test_compute_model_performance_multitask_classifier():
    n_data_points = 20
    n_features = 1
    n_tasks = 2
    n_classes = 2

    X = np.ones(shape=(n_data_points // 2, n_features)) * -1
    X1 = np.ones(shape=(n_data_points // 2, n_features))
    X = np.concatenate((X, X1))
    class_1 = np.array([[0.0, 1.0] for x in range(int(n_data_points / 2))])
    class_0 = np.array([[1.0, 0.0] for x in range(int(n_data_points / 2))])
    y1 = np.concatenate((class_0, class_1))
    y2 = np.concatenate((class_1, class_0))
    y = np.stack([y1, y2], axis=1)
    dataset = NumpyDataset(X, y)

    features = layers.Input(shape=(n_features))
    dense = layers.Dense(n_tasks * n_classes)(features)
    logits = layers.Reshape((n_tasks, n_classes))(dense)
    output = layers.Softmax()(logits)
    keras_model = tf.keras.Model(inputs=features, outputs=[output, logits])
    model = dc.models.KerasModel(keras_model,
                                 dc.models.losses.SoftmaxCrossEntropy(),
                                 output_types=['prediction', 'loss'],
                                 learning_rate=0.01,
                                 batch_size=n_data_points)

    model.fit(dataset, nb_epoch=1000)
    metric = dc.metrics.Metric(dc.metrics.roc_auc_score,
                               np.mean,
                               mode="classification")

    scores = model.evaluate_generator(model.default_generator(dataset),
                                      [metric],
                                      per_task_metrics=True)
    scores = list(scores[1].values())
    # Loosening atol to see if tests stop failing sporadically
    assert np.all(np.isclose(scores, [1.0, 1.0], atol=0.50))


@pytest.mark.tensorflow
def test_compute_model_performance_singletask_classifier():
    """Computes model performance on singletask dataset with one-hot label encoding."""
    n_data_points = 20
    n_features = 10

    X = np.ones(shape=(int(n_data_points / 2), n_features)) * -1
    X1 = np.ones(shape=(int(n_data_points / 2), n_features))
    X = np.concatenate((X, X1))
    class_1 = np.array([[0.0, 1.0] for x in range(int(n_data_points / 2))])
    class_0 = np.array([[1.0, 0.0] for x in range(int(n_data_points / 2))])
    y = np.concatenate((class_0, class_1))
    dataset = NumpyDataset(X, y)

    features = layers.Input(shape=(n_features,))
    dense = layers.Dense(2)(features)
    output = layers.Softmax()(dense)
    keras_model = tf.keras.Model(inputs=features, outputs=[output])
    model = dc.models.KerasModel(keras_model,
                                 dc.models.losses.SoftmaxCrossEntropy(),
                                 learning_rate=0.1)

    model.fit(dataset, nb_epoch=1000)
    metric = dc.metrics.Metric(dc.metrics.roc_auc_score,
                               np.mean,
                               mode="classification",
                               n_tasks=1)

    scores = model.evaluate_generator(model.default_generator(dataset),
                                      [metric],
                                      per_task_metrics=True)
    scores = list(scores[1].values())
    assert np.isclose(scores, [1.0], atol=0.05)


@pytest.mark.tensorflow
def test_compute_model_performance_multitask_regressor():
    random_seed = 42
    n_data_points = 20
    n_features = 2
    n_tasks = 2
    np.random.seed(seed=random_seed)

    X = np.random.rand(n_data_points, n_features)
    y1 = np.array([0.5 for x in range(n_data_points)])
    y2 = np.array([-0.5 for x in range(n_data_points)])
    y = np.stack([y1, y2], axis=1)
    dataset = NumpyDataset(X, y)

    features = layers.Input(shape=(n_features,))
    dense = layers.Dense(n_tasks)(features)
    keras_model = tf.keras.Model(inputs=features, outputs=[dense])
    model = dc.models.KerasModel(keras_model,
                                 dc.models.losses.L2Loss(),
                                 learning_rate=0.1)

    model.fit(dataset, nb_epoch=1000)
    metric = [
        dc.metrics.Metric(dc.metrics.mean_absolute_error,
                          np.mean,
                          mode="regression"),
    ]
    scores = model.evaluate_generator(model.default_generator(dataset),
                                      metric,
                                      per_task_metrics=True)
    scores = list(scores[1].values())
    assert np.all(np.isclose(scores, [0.0, 0.0], atol=1.0))
